def frequent_words(text,k):
    """Find the most frequent k-mers in the gene sequence. K-mer is a short sequence of DNA that is 'k' bases long, usually the length needed to encode a protein for some biological process.

    Args:
    text: string. The sequence of DNA where we are trying to find frequent k-mers.
    k: int. The number of bases in our k-mer- our substring of DNA that encodes a protein.

    Returns:
    frequent_words_list: list. A list of the most frequent k-mers in our gene sequence.
    """
    words = {}
    maxOccurrences = 0
    frequent_words_list = []
    for i in range(0, len(text)-k+1):
        key = text[i:i+k]
        if key in words:
            words[key] += 1
            if words[key] > maxOccurrences:
                maxOccurrences = words[key]
        else:
            words[key] = 1
            if(maxOccurrences <= 1):
                maxOccurrences = 1
    for key in words:
        if(words[key] == maxOccurrences):
            frequent_words_list.append(key)
    return frequent_words_list

def reverse_complement(pattern):
    """Find the reverse complement of this DNA sequence. Useful to see what the complementary strand should look like after replication (if ignoring the effect of mutations)
    
    Args:
    pattern: string. The sequence of DNA we want to find the complement of.

    Returns:
    reverse_complement: string. The reverse complement of the input DNA sequence, according to base pairing rules.
    """
    reverse_complement = ''
    for i in range(0,len(pattern)):                         #can we vectorize?
        current = pattern[len(pattern)-i-1]
        if(current == 'A'):
            reverse_complement += 'T'
        elif(current == 'T'):
            reverse_complement += 'A'
        elif(current == 'G'):
            reverse_complement += 'C'
        else:
            reverse_complement += 'G'
    return reverse_complement

def pattern_indices(pattern,gene_sequence):
    """ Find the indices within the gene_sequence where pattern begins. """
    indices = []
    pattern_seen = False
    pattern_start_index = 0
    for i in range(0,len(gene_sequence)-len(pattern)+1):
        tmp = gene_sequence[i:i+len(pattern)]
        if(tmp == pattern):
            indices.append(i) 
    return indices

def pattern_count(text,pattern):
    """Find the number of occurrences of the substring 'pattern' in larger string 'text'."""
    count = 0
    for i in range(0,len(text) - len(pattern) + 1):
        if(text[i:i+len(pattern)] == pattern):
            count += 1
    return count

def clump_finder(gene_sequence,k,l,t):
    """Find the k-mers (clumps) that occur at least t times within a window of the DNA that is l bases long.

    Args:
    gene_sequence: string. Entire genome or subset of the genome
    k: int. Length of our k-mer.
    L: int. Length of our 'window'.
    t: int. Number of occurrences of this clump we expect.

    Returns: 
    clumps: list. All k-mers that match our conditions. 
    """
    k_mers = {}                             
    debug = False
    clumps = set([])
    #populate our dictionary of all k-mers
    for i in range(0,len(gene_sequence)-k+1):      
        k_mer = gene_sequence[i:i+k]
        if k_mer in k_mers:
            k_mers[k_mer].append(i)
        else:
            k_mers[k_mer] = [i]

    #find k_mers that occur at least t times in L window
    for k_mer in k_mers:
        indices_list = k_mers[k_mer]
        for i in range(0,len(indices_list)-t+1):
            first_index = indices_list[i]
            last_index = indices_list[i+t-1]+k-1
            if(last_index - first_index < l):
                clumps.add(k_mer)
                break

    return clumps

def skew(gene_sequence,end_index):
    """ Calculate the difference between the quantity of G bases and the quantity of C bases. Helps to determine the effect of deamination and where 'ori' might lie.

    Args:
    gene_sequence: string. The DNA string on which to calculate the skew. Will most likely be the entire genome.
    end_index: int. The stopping point for calculating the skew.

    Returns: 
    skew: int. The difference between G and C bases in this gene_sequence.
    """
    return gene_sequence[0:end_index].count('G')-gene_sequence[0:end_index].count('C')


class Genome():

    def __init__(self,genome,description="",source_file=""):
        self.genome = genome
        self.description = description
        self.source_file = source_file

    def __repr__(self):
        return """==================== Genome ======================\n\n{}\n\nsource: {}\n\ncontains {} nucleotides\n==================================================""".format(self.description,self.source_file,len(self.genome))

    def minimum_skew(self):
        """Find the point in our genome where the skew is minimized- will help in finding 'ori'.

        Args:
        self.genome: string. Our genome.

        Returns:
        min_indices: list. A list of indices where the skew is at a minimum in our genome.
        """
        min_indices = []
        skew_cache = [0]
        minimum_seen = 0
        min_indices = [0]
        for i in range(1,len(self.genome)+1):
            _skew = skew_cache[i-1]+skew(self.genome[i-1:i],1)
            skew_cache.append(_skew)
            if _skew > minimum_seen:
                continue
            elif _skew == minimum_seen:
                min_indices.append(i)
            else:
                min_indices.clear()
                minimum_seen = _skew
                min_indices.append(i)
        return min_indices



# from genome import Genome
# from genome import skew
from genome import *

def main():
    clumps = clump_finder("""ACGTACGT""",1,5,2)
    print("clumps: {}".format(clumps))

# def main():
#     gen = open('dataset_7_6.txt', 'r').read()
#     genome = Genome(gen,'The genome of E.Coli','dataset_7_6.txt')
#     min_skew = genome.minimum_skew()
#     _skew = skew(gen,5)
#     print("minimum skew: \n\n{}\n".format(min_skew))
#     print(genome)
#     print("\nskew: {}\n".format(_skew))


if __name__ == '__main__':
    main()



import sys,traceback

def PatternCount(Text,pattern):
    count = 0
    for i in range(0,len(Text) - len(pattern) + 1):
        if(Text[i:i+len(pattern)] == pattern):
            count += 1
    return count

def FrequentWords(Text,k):
    words = {}
    maxOccurrences = 0
    frequent_words_list = []
    for i in range(0, len(Text)-k+1):
        key = Text[i:i+k]
        if key in words:
            words[key] += 1
            if words[key] > maxOccurrences:
                maxOccurrences = words[key]
        else:
            words[key] = 1
            if(maxOccurrences <= 1):
                maxOccurrences = 1
    for key in words:
        if(words[key] == maxOccurrences):
            frequent_words_list.append(key)
    return frequent_words_list

def ReverseComplement(Pattern):
    reverseComplement = ''
    for i in range(0,len(Pattern)):
        current = Pattern[len(Pattern)-i-1]
        if(current == 'A'):
            reverseComplement += 'T'
        elif(current == 'T'):
            reverseComplement += 'A'
        elif(current == 'G'):
            reverseComplement += 'C'
        else:
            reverseComplement += 'G'
    return reverseComplement

def PatternIndices(Pattern,Genome):
    indices = []
    pattern_seen = False
    pattern_start_index = 0
    for i in range(0,len(Genome)-len(Pattern)+1):
        tmp = Genome[i:i+len(Pattern)]
        if(tmp == Pattern):
            indices.append(i) 
    return indices

# def ClumpFinder(genome,k,L,t):
#     i = 0
#     debug = False
#     tmp_dict = {}
#     clumps = set([])
#     current_window = ''
#     while i <= len(genome)-L:    
#         current_window = genome[i:i+L]
#         if debug:
#             print('current_window: {}'.format(current_window))
#             input('press enter')
#         for j in range(0,len(current_window)-k+1):
#             k_mer = current_window[j:j+k]
#             if debug:
#                 print('k_mer: {}'.format(k_mer))
#                 input('press enter')
#             if k_mer in tmp_dict:
#                 tmp_dict[k_mer] += 1
#             else:
#                 tmp_dict[k_mer] = 1
#         for key in tmp_dict:
#             if tmp_dict[key] == t:
#                 clumps.add(key)
#         tmp_dict = {}
#         i+=1 
#     return clumps

def ClumpFinder(genome,k,L,t):
    k_mers = {}                             #our dictionary holding all kmers, with list of start indices
    debug = False
    clumps = set([])
    for i in range(0,len(genome)-k+1):      #populate our dictionary
        k_mer = genome[i:i+k]
        if k_mer in k_mers:
            k_mers[k_mer].append(i)
        else:
            k_mers[k_mer] = [i]
    if debug:
        print('kmers: \n\n{}'.format(k_mers))
        input('press enter')
    for k_mer in k_mers:
        indices_list = k_mers[k_mer]
        # if debug:
        #     print('k_mer: {} \t indices_list: {}'.format(k_mer,indices_list))
        #     input('press enter')
        for i in range(0,len(indices_list)-t+1):
            first_index = indices_list[i]
            last_index = indices_list[i+t-1]+k-1
            if(last_index - first_index < L):
                clumps.add(k_mer)
                if debug:
                    print('k_mer "{}" forms a clump between {} and {}'.format(k_mer,first_occurrence,t_occurrence))
                    print('\nindices_list: {}'.format(indices_list))
                    input('press enter')
                break

    return clumps





f = open('E_coli.txt','r')
genome = f.read()
clumps = ClumpFinder(genome,9,500,3)
tmp = ''
for clump in clumps:
    tmp = tmp + ' ' + str(clump) 
print(tmp)
print("\n\n"+str(len(clumps)) + "\n")





import time

def skew(genome,i):
    return genome[0:i].count('G')-genome[0:i].count('C')

def minimum_skew(genome):
    min_indices = []
    skew_cache = [0]
    minimum_seen = 0
    min_indices = [0]
    for i in range(1,len(genome)):
        _skew = skew_cache[i-1]+skew(genome[i-1:i],1)
        skew_cache.append(_skew)
        if _skew > minimum_seen:
            continue
        elif _skew == minimum_seen:
            min_indices.append(i)
        else:
            min_indices.clear()
            minimum_seen = _skew
            min_indices.append(i)
    return min_indices
    
##
##start_time = time.time()
##for i in range(0,15):
##    print('{} '.format(skew('GAGCCACCGCGATA',i)))
##exc_time = time.time()-start_time
##print("\n\nexecution time: {} \n".format(exc_time))

f = open('dataset_7_6.txt','r')
genome = f.read()
start_time = time.time()
print("\nMinimum skew indices: {}".format(minimum_skew(genome)))
exc_time = time.time()-start_time
print("\n\nexecution time: {} \n".format(exc_time))
